# This module is part of the Pyndex project and is Copyright 2003 Amir
# Bakhtiar (amir@divmod.org). This is free software; you can redistribute
# it and/or modify it under the terms of version 2.1 of the GNU Lesser
# General Public License as published by the Free Software Foundation.

"""
train(pool, obj)
guess(obj) -> pool

To use this class override the getTokens method
to return a list of strings that will be the
'features' of your object that Bayes trains on.
"""

from pyndex.indexer import Index
from pyndex.analysis import TextSplitter
from pyndex import __version__
import operator
import math
from types import StringType

class Pool(dict):
    __slots__ = ['training']

    def trainedOn(self, id):
        return id in self.training
        
        
class Bayes(Index):
    __name_prefix__ = 'pyndex.bayes.'

        
    def __init__(self, db=None, a=None):
        if db is None:
            self.db = self.getMemDb()
        else:
            self.db = db
            
        if a is None:
            self.analyzer = TextSplitter()
        else:
            self.analyzer = a

        self.guesser = self.robinson
        
        self.setupDB()

        self.corpcache = Pool()
        self.corpcache.training = None
        
        self.cache = {}
        self.dirty = True

    def setupDB(self):
        pref = self.__name_prefix__
        
        fschema = pref + "files[docid:I,docname:S,wordcount:I,deleted:I,words[word:S,count:I],training[int:I,str:S]]"
        self.files = self.db.getas(fschema)
        
        sschema = pref + "system[key:S,int:I,str:S]"
        self.system = self.db.getas(sschema)
        
        self.ruid = self.getSystem(key='uid', int=0, str='montonically increasing integer ')
        self.rwc = self.getSystem(key='wordcount', int=0, str='tot non-unique words')
        self.getSystem(key='fschema', int=0, str=fschema)
        self.getSystem(key='version', int=0, str=__version__)
        
        self.corpus, ignore = self.getFile('__corpus')
        self.rGuess = self.getSystem(key='guesses', int=0, str='total guesses')
        self.rTrain = self.getSystem(key='training', int=0, str='total training')

    def getTokens(self, obj):
        if isinstance(obj, StringType):
            return self.split(obj)
        else:
            return ''

    def train(self, poolName, text, id=None, commit=True):
        """Add text to the index.

        docId is a string identifier with which the text is associated.
        This is usually a unique string used to identify an indexed file.

        docId is the value that is returned as the result of a
        seach using Index.find()
        """

        words = self.getTokens(text)
        wc = len(words)

        self._train(poolName, words)
        
        self.rTrain.int += 1
        self.dirty = True

        if id:
            pool, ignore = self.getFile(poolName)
            if isinstance(id, ''.__class__):
                pool.training.append(str=id)
            else:
                pool.training.append(int=id)

        if commit:
            self.commit()

    def _train(self, poolName, tokens):
        # tokens should be a list of strings
        doc, ignore = self.getFile(poolName)
        wc = len(tokens)
        
        doc.wordcount += wc
        self.corpus.wordcount += wc
        self.updateCount(wc)
        for token in tokens:
            self.setDocWords(doc, token)
            self.setDocWords(self.corpus, token)

    def trainedOn(self, id):            
        for p in self.cache.values():
            if id in p.training:
                return True
        return False

    def guess(self, obj):
        tokens = self.getTokens(obj)
        pools = self.poolProbs()

        res = {}
        for pname, pprobs in pools.items():
            p = self.getProbs(pprobs, tokens)
            if len(p) != 0:
                res[pname]=self.guesser(p, pname)
        res = res.items()
        res.sort(lambda x,y: cmp(y[1], x[1]))
        self.rGuess.int += 1
        return res

    def totalGuesses(self):
        return self.rGuess.int
    
    def totalTraining(self):
        return self.rTrain.int
        
    def poolProbs(self):
        if self.dirty:
            for w in self.corpus.words:
                self.corpcache[w.word]=w.count
            self.buildCache()
            self.dirty = False
        return self.cache
    
    def buildCache(self):
        """ merges good and bad corpora and computes probabilities
        """
        
        for p in self.files:
            # skip our special pool
            if p.docname in ['__corpus']:
                continue
            
            pool = self.cache.setdefault(p.docname, Pool())
            pool.training = [r.int for r in p.training]

            poolCount = len(p.words)
            themCount = max(len(self.corpus.words) - poolCount, 1)

            for word, totCount in self.corpcache.items():
                # for every word in the copus
                # check to see if this pool contains this word
                row, exists = p.words.locate(word=word)
                if exists:
                    # Let's use the count from the db
                    thisCount = float(p.words[row].count)
                else:
                    # Let's default a value
                    thisCount = 0.0

                otherCount = float(totCount) - thisCount
                
                goodMetric = min(1.0, otherCount/poolCount)
                badMetric = min(1.0, thisCount/themCount)
                f = badMetric / (goodMetric + badMetric)
                
                #print w, g, ngood, b, nbad, f
                # PROBABILITY_THRESHOLD
                if abs(f-0.5) >= 0.1 :
                    # GOOD_PROB, BAD_PROB
                    pool[word] = max(0.0001, min(0.9999, f))


    def getProbs(self, pool, words):
        """ extracts the probabilities of tokens in a message
        """
    
        #probs = [pool.get(word, 0.45) for word in words]
        #probs = [(word, pool[word]) for word in words if word in pool]
        #probs = [(word,pool.get(word, 0.0001)) for word in words]  
        probs = [(word, pool[word]) for word in words if word in pool]
        #probs.sort(lambda x,y: cmp(abs(y[1]-0.5),abs(x[1]-0.5)) or cmp(x[1],y[1]))
        probs.sort(lambda x,y: cmp(y[1],x[1]))
        return probs[:2048]


    def robinson(self, probs, ignore):
        """ computes the probability of a message being spam (Robinson's method)
            P = 1 - prod(1-p)^(1/n)
            Q = 1 - prod(p)^(1/n)
            S = (1 + (P-Q)/(P+Q)) / 2
            Courtesy of http://christophe.delord.free.fr/en/index.html
        """
        
        nth = 1./len(probs)
        P = 1.0 - reduce(operator.mul, map(lambda p: 1.0-p[1], probs), 1.0) ** nth
        Q = 1.0 - reduce(operator.mul, map(lambda p: p[1], probs)) ** nth
        S = (P - Q) / (P + Q)
        return (1 + S) / 2


    def robinsonFisher(self, probs, ignore):
        """ computes the probability of a message being spam (Robinson-Fisher method)
            H = C-1( -2.ln(prod(p)), 2*n )
            S = C-1( -2.ln(prod(1-p)), 2*n )
            I = (1 + H - S) / 2
            Courtesy of http://christophe.delord.free.fr/en/index.html
        """
        n = len(probs)
        try: H = chi2P(-2.0 * math.log(reduce(operator.mul, map(lambda p: p[1], probs), 1.0)), 2*n)
        except OverflowError: H = 0.0
        try: S = chi2P(-2.0 * math.log(reduce(operator.mul, map(lambda p: 1.0-p[1], probs), 1.0)), 2*n)
        except OverflowError: S = 0.0
        return (1 + H - S) / 2
    

    def popFile(self, probs, pName):
        pool, ignore = self.getFile(pName)
        myCount = float(pool.wordcount)

        sum = reduce(operator.add, [math.log(p[1]) for p in probs])

        sum += math.log(float(myCount)/len(self.corpcache))
        return sum


    
def chi2P(chi, df):
    """ return P(chisq >= chi, with df degre of freedom)

    df must be even
    """
    assert df & 1 == 0
    m = chi / 2.0
    sum = term = math.exp(-m)
    for i in range(1, df/2):
        term *= m/i
        sum += term
    return min(sum, 1.0)
