# This module is part of the Pyndex project and is Copyright 2003 Amir
# Bakhtiar (amir@divmod.org). This is free software; you can redistribute
# it and/or modify it under the terms of version 2.1 of the GNU Lesser
# General Public License as published by the Free Software Foundation.

"""Indexer is a simple full text indexer implementing an inverted index.
It uses Metakit as its storage mechanism.
It offers Porter stemming.
"""

import string
import metakit
from pyndex.analysis import TextSplitter
from pyndex.search import Query
from pyndex.analysis import DefaultAnalyzer
from pyndex import __version__
import types
import math


class Index(object):
    __slots__ = ['db','words','files','system','ruid', 'docWords',
                 'rwc', 'offsets', 'analyzer', 'wordCache']
    __name_prefix__ = 'pyndex.'
    
    def __init__(self, database=None, analyzer=None, offsets = True, docWords = False, create=False):
        """Create a new index.

        If there is a name, it is assumed to be the filename for the
        index to be stored in.

        If name is None, an in-memory index is created.

        If porter is set to True, Porter stemming is used.
        Porter stemming strips words of common ending such as '-ed' and '-ing'.
        When using Porter stemming the words 'ambushes', 'ambushed' and 'ambushing'
        all get stemmed and index as 'ambush'.

        This indexer is not case-sensitive since Metakit text fields are
        not text sensitive
        """

        # create is really a stand in for 'not read-only'

        if database is None:
            self.db = self.getMemDb()
        else:
            if isinstance(database, types.StringType):
                self.db = metakit.storage(database, create)
            else:
                self.db = database
            
        if analyzer is None:
            self.analyzer = DefaultAnalyzer()
        else:
            self.analyzer = analyzer

        self.offsets = offsets
        self.docWords = docWords

        pref = self.__name_prefix__

        if offsets:
            wschema = pref + "words[word:S,index[docid:I,count:I,offsets[offset:I]]]"
        else:
            wschema = pref + "words[word:S,index[docid:I,count:I]]"
        self.words = self.db.getas(wschema)
        
        if docWords:
            fschema = pref + "files[docid:I,docname:S,wordcount:I,hi:I,norm:I,deleted:I,words[word:S,count:I]]"
        else:
            fschema = pref + "files[docid:I,docname:S,wordcount:I,hi:I,norm:I,deleted:I]"
        self.files = self.db.getas(fschema)
        
        sschema = pref + "system[key:S,int:I,str:S]"
        self.system = self.db.getas(sschema)
        self.rwc = self.getSystem(key='wordcount', int=0, str='tot non-unique words')
        self.ruid = self.getSystem(key='uid', int=0, str='montonically increasing integer ')
        self.getSystem(key='wschema', int=0, str=wschema)
        self.getSystem(key='fschema', int=0, str=fschema)
        self.getSystem(key='version', int=0, str=__version__)
        self.wordCache = {}

        
    def getMemDb(self):
        """Create an in-memory database."""
        newdb = metakit.storage()
        return newdb
    

    def wordCount(self):
        return self.rwc.int
    

    def peekUid(self):
        return self.ruid.int
        

    def nextUid(self):
        r =  self.ruid.int
        self.ruid.int += 1
        return r
        

    def commit(self):
        self.wordCache={}
        self.db.commit()
        

    def close(self):
        self.commit()
        

    def split(self, text):
        return self.analyzer.split(text)


    def updateCount(self, delta):
        self.rwc.int += delta
        
        
    def getFile(self, dId, create = True):
        # Find the row for this document, create one if needed
        # Naught shortcut for now
        # if dId is a string, we search by filename
        # if dId is not a string, we assume it is and int and search docid
        
        # Todo add caching
    
        if isinstance(dId, types.StringType):
            row = self.files.find(docname=dId)
            if row == -1:
                exists = False
                # no uid provided - generate our own
                nid = self.nextUid()
                dname = dId
                row = self.files.append(docid=nid, docname=dname, wordcount=0, deleted=0)
            else:
                exists = True
        else:
            # dId is an int and we will use it as our id
            # the docname defaults to the string of the id
            row, exists = self.files.locate(docid=dId)
            if not exists:
                nid = dId
                dname = str(nid)
                # insert
                if create:
                    self.files.insert(row, docid=nid, docname=dname, wordcount=0, deleted=0)
                    #return (self.files[row], exists)
                else:
                    return (None, False)
        return (self.files[row], exists)
    

    def getWord(self, word, create = True):
        # Todo add caching
        row, exists = self.words.locate(word=word)
        if not exists:
            # insert
            if create:
                self.words.insert(row, word=word)
            else:
                return None
        return self.words[row]

    def getSystem(self, key, int = 0, str='', create = True):
        # Find the row for this word, create one if needed
        row = self.system.find(key=key)
        if row == -1:
            # append
            if create:
                row = self.system.append(key=key, int = int, str=str)
            else:
                return None
        return self.system[row]
    

    def setDocWords(self, doc, word):
        """Update the words for this document
        e.g. list of unique words in this doc + their freq"""
        
        # Find the row for this file, create one if needed
        dwRow, exists = doc.words.locate(word=word)
        if not exists:
            # insert
            doc.words.insert(dwRow, word=word, count=1)
        else:
            # Increment the word count for this word in this document 
            (doc.words[dwRow]).count += 1


    def index(self, documentId, text, start = 0):
        """Add text to the index.

        docId is a string identifier with which the text is associated.
        This is usually a unique string used to identify an indexed file.

        docId is the value that is returned as the result of a
        seach using Index.find()
        """

        words = self.split(text)
        wc = len(words)

        doc, exists = self.getFile(documentId)
        if not exists:
            # self.files[row].wordcount += wc
            doc.wordcount += wc
            self.updateCount(wc)
            doc.norm = self.norm(wc)
        
        pos = start
        max = 0
        for word in words:
            entry = self.getWord(word)
            
            # Find the row for this file, create one if needed
            fRow, exists = entry.index.locate(docid=doc.docid)
            if not exists:
                # insert
                entry.index.insert(fRow, docid=doc.docid, count=1)
            else:
                # Increment the word count for this word in this document 
                (entry.index[fRow]).count += 1
            if entry.index[fRow].count > max:
                max = entry.index[fRow].count

            if self.docWords is True:
                self.setDocWords(doc, word)

            if self.offsets is True:
                # Record the offset of this word in this document
                (entry.index[fRow]).offsets.append(offset=pos)
                pos += 1
        if max > doc.hi:
            doc.hi = max
        return (doc, exists)
        

    def find(self, qString):
        """Find documents in the index that include the words in queryText.
        qString can be a word , a list of words or a phrase.
        when it is a list of words find 'ands' all the word
        The result is a dictionary in the form of:
            {'word': [('docid', occur, wordcount), ('docid2', occur, wordcount)],
            'word2': [('docid', occur, wordcount)]}
        Where word is the word that was searched for, docid is the document it
        appears in, occur is the number of times that word occurs in docid and
        wordcount is the total number of words in docid.

        If queryText includes more than one word, each word gets its own entry
        in the result dictionary.
        """
        
        q = Query(self)
        return q.find(qString)

    def write(self, fname):
        fp = open(fname,'wb')
        self.db.save(fp)
        fp.close()

    def optimize(self):
        pass

    def norm(self, wc):
        """Computes the normalization byte for a document given the total number of
        terms contained in the document.  These values are stored in an index and
        used by the search code

        Scales 1/sqrt(numTerms) into a byte, i.e. 256/sqrt(numTerms).
        Math.ceil is used to ensure that even very long documents don't get a
        zero norm byte, as that is reserved for zero-lengthed documents and
        deleted documents."""

        if wc == 0:
            return 0.0
        return int((math.ceil(255.0 / math.sqrt(wc)))) & 0xFF
