# This module is part of the Pyndex project and is Copyright 2003 Amir
# Bakhtiar (amir@divmod.org). This is free software; you can redistribute
# it and/or modify it under the terms of version 2.1 of the GNU Lesser
# General Public License as published by the Free Software Foundation.

import math
import metakit

class PhrasePosition(object):
    __slots__ = ['count', 'docid', 'offset', 'position', 'word']
    
    
    def __init__(self, w, d, o, pos):
        self.word = w
        self.docid = d
        self.offset = o
        self.position = pos
        

    def __cmp__(self, other):
        if self.docid != other.docid:
            return cmp(self.docid, other.docid)
        else:
            return cmp(self.offset, other.offset)
        

    def __repr__(self):
        return ('<pp:' + self.word + ' ' +
                self.docid + ' ' +
                str(self.offset) + ' ' +
                str(self.position) + '>')
    
        

class Hit(object):
    __slots__ = ['doc', 'score', 'count', 'numDocs']

    def __init__(self, doc, count, numDocs, corpCount):
        self.count = count
        self.doc = doc
        self.numDocs = numDocs
        self._score(corpCount)

    def __cmp__(self, other):
        return cmp(other.score, self.score)

    def __repr__(self):
        return ('<Hit:docid ' + str(self.doc.docid) +
                ', score ' + str(self.score) + '>')

    def _score(self, wc):
        dk = self.numDocs
        i = idf(dk, wc)
        tf = termFreq(self.count)
        self.score = tf * i

    def _score(self, wc):
        tf = termFreq(self.count)
        self.score = tf * (self.doc.norm/255.0)                    
    

class Query(object):
    def __init__(self, db):
        self.db = db
        self.corpusCount = len(self.db.words)
    
    def find(self, qString):
        if len(self.db.words) is 0:
            "Empty db"
            return []
        if qString.startswith('"'):
            qWords = qString.strip('"').split()
            return self.findPhrase(qWords)
        else:
            qWords = qString.split()
            if len(qWords) == 1:
                f = self.findOne(qWords[0])
                return f
            else:
                f = self.findAll(qWords)
                return f

    def findOne(self, qWord):
        found = self.db.getWord(qWord, False)
        if found is None:
            return []

        dk = len(found.index)
        r = []
        for d in found.index:
            doc, ignore = self.db.getFile(d.docid, False)
            r.append(Hit(doc, len(d.offsets), dk, self.corpusCount))
        r.sort()
        return r

    def findAny(self, qWords):
        pass

    def findAll(self, qWords):
        q1 = qWords[0]
        first = self.db.getWord(q1, False)
        if first is None:
            return []
        
        qWords = qWords[1:]

        # Needed for the join e.g. which column to join on
        docIdProp = metakit.property('I', 'docid')

        wordCache = {}
        wordCache[first.word] = first
        # seed the result
        resultIndex = first.index
        for q in qWords:
            # iteratively join all word indexes
            w = self.db.getWord(q, False)
            if w is not None:
                wordCache[q] = w
                resultIndex = resultIndex.join(w.index, docIdProp)
            else:
                return []

        # Scoring
        res = {}
        for r in resultIndex:
            # for each document
            res[r] = 0
            doc, ignore = self.db.getFile(r.docid, False)
            for word, row in wordCache.items():
                # for each word
                rd = row.index[row.index.find(docid=r.docid)]
                res[r] += len(rd.offsets)

        r = []
        for d, count in res.items():
            doc, ignore = self.db.getFile(d.docid, False)
            h = Hit(doc, count, len(res), self.corpusCount)
            r.append(h)
        r.sort()
        return r


    def getPhrasePos(self, qWords):
        all = self.findAll(qWords)
        w = self.db.words
        positions = []
        pos = 0
        for word in qWords:
            wRow = w[w.search(word=word)]
            for h in all:
                d = h.doc.docid
                iRow = wRow.index[wRow.index.search(docid=d)]
                for oRow in iRow.offsets:
                    o = oRow.offset
                    positions.append(PhrasePosition(word, d, o, pos))
            pos += 1
        positions.sort()
        return positions
        

    def findPhrase(self, qWords):
        positions = self.getPhrasePos(qWords)
        
        w = self.db.words
        s = 1
        res = []
        seen = {}
        p = PhrasePosition('','',None,0)
        while s < len(positions):
            s, r = self.findPos(positions, s-1, p, len(qWords))
            if r is not None:
                did = r[0].docid
                doc, ignore = self.db.getFile(did, create=False)
                if did not in seen:
                    seen[did] = (doc, 1)
                else:
                    dc, ct = seen[did]
                    seen[did] = (dc, ct + 1)
        r = []
        for doc, count in seen.values():
            r.append(Hit(doc, count, len(seen), self.corpusCount))
        r.sort()
        return r


    def findPos(self, l, s, pp, num):
        if pp.offset is None:
            # find head
            s, r = self.seek(l,s,pp)
            if r is None:
                # not found
                return s, None
            else:
                # Found, let's start the search proper
                p = PhrasePosition('', r.docid, (r.offset + 1), (r.position + 1))
                return self.findPos(l, s, p, num - 1)
        else:
            res = []
            for i in range(num):
                p = PhrasePosition('', pp.docid, pp.offset + i, pp.position + i)
                s, r = self.seek(l,s,p)
                if r is None:
                    return s, None
                else:
                    res.append(r)
            return s, res
        
            
    def seek(self, l, s, pp):
        while s < len(l):
            i = l[s]
            s += 1
            if pp.offset is not None:
                if (pp.docid != i.docid or i.offset > pp.offset):
                    # Past the desired spot
                    return s, None
            if i.position == pp.position:
                # semi-hit
                if pp.offset is not None and pp.offset == i.offset:
                    # Real hit
                    return s, i
                else:
                    # found start
                    return s, i
        return s, None
        

def idf(docFreq, numDocs):
    return math.log(numDocs/(docFreq + 1.0)) + 1.0

def termFreq(freq):
    return math.sqrt(freq)

        
